<?php
/*
 * Copyright (C) 2017 beGateway
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * @author      beGateway
 * @copyright   2017 beGateway
 * @license     http://opensource.org/licenses/gpl-2.0.php GNU General Public License, version 2 (GPL-2.0)
 */

namespace BeGateway\BeGateway\Model\Method;

/**
 * Checkout Payment Method Model Class
 * Class Checkout
 * @package BeGateway\BeGateway\Model\Method
 */
class Checkout extends \Magento\Payment\Model\Method\AbstractMethod
{
    use \BeGateway\BeGateway\Model\Traits\OnlinePaymentMethod;
    use \BeGateway\BeGateway\Model\Traits\Logger;

    const CODE = 'begateway_checkout';
    /**
     * Checkout Method Code
     */
    protected $_code = self::CODE;

    protected $_canOrder                    = true;
    protected $_isGateway                   = true;
    protected $_canCapture                  = true;
    protected $_canCapturePartial           = true;
    protected $_canRefund                   = true;
    protected $_canCancelInvoice            = true;
    protected $_canVoid                     = true;
    protected $_canRefundInvoicePartial     = true;
    protected $_canAuthorize                = true;
    protected $_isInitializeNeeded          = false;

    /**
     * Get Instance of the Magento Code Logger
     * @return \Monolog\Logger
     */
    protected function getLogger()
    {
        return $this->_logger;
    }

    /**
     * Checkout constructor.
     * @param \Magento\Framework\Model\Context $context
     * @param \Magento\Framework\App\Action\Context $actionContext
     * @param \Magento\Framework\Registry $registry
     * @param \Magento\Framework\Api\ExtensionAttributesFactory $extensionFactory
     * @param \Magento\Framework\Api\AttributeValueFactory $customAttributeFactory
     * @param \Magento\Payment\Helper\Data $paymentData
     * @param \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig
     * @param \Magento\Payment\Model\Method\Logger $logger
     * @param \Magento\Store\Model\StoreManagerInterface $storeManager
     * @param \Magento\Checkout\Model\Session $checkoutSession
     * @param \BeGateway\BeGateway\Helper\Data $moduleHelper
     * @param \Magento\Framework\Model\ResourceModel\AbstractResource|null $resource
     * @param \Magento\Framework\Data\Collection\AbstractDb|null $resourceCollection
     * @param array $data
     */
    public function __construct(
        \Magento\Framework\Model\Context $context,
        \Magento\Framework\App\Action\Context $actionContext,
        \Magento\Framework\Registry $registry,
        \Magento\Framework\Api\ExtensionAttributesFactory $extensionFactory,
        \Magento\Framework\Api\AttributeValueFactory $customAttributeFactory,
        \Magento\Payment\Helper\Data $paymentData,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Magento\Payment\Model\Method\Logger  $logger,
        \Magento\Store\Model\StoreManagerInterface $storeManager,
        \Magento\Checkout\Model\Session $checkoutSession,
        \BeGateway\BeGateway\Helper\Data $moduleHelper,
        \Magento\Framework\Model\ResourceModel\AbstractResource $resource = null,
        \Magento\Framework\Data\Collection\AbstractDb $resourceCollection = null,
        array $data = []
    ) {
        parent::__construct(
            $context,
            $registry,
            $extensionFactory,
            $customAttributeFactory,
            $paymentData,
            $scopeConfig,
            $logger,
            $resource,
            $resourceCollection,
            $data
        );
        $this->_actionContext = $actionContext;
        $this->_storeManager = $storeManager;
        $this->_checkoutSession = $checkoutSession;
        $this->_moduleHelper = $moduleHelper;

        $this->_logger = $logger;

        $this->_configHelper =
            $this->getModuleHelper()->getMethodConfig(
                $this->getCode()
            );
    }

    /**
     * Get Default Payment Action On Payment Complete Action
     * @return string
     */
    public function getConfigPaymentAction()
    {
        return \Magento\Payment\Model\Method\AbstractMethod::ACTION_ORDER;
    }

    /**
     * Get Available Checkout Transaction Types
     * @return array
     */
    public function getCheckoutTransactionTypes()
    {
        $selected_types = $this->getConfigHelper()->getTransactionTypes();

        return $selected_types;
    }

    /**
     * Get Available Checkout Payment Method Types
     * @return array
     */
    public function getCheckoutPaymentMethodTypes()
    {
        $selected_types = $this->getConfigHelper()->getPaymentMethodTypes();

        return $selected_types;
    }

    /**
     * Create a Web-Payment Form Instance
     * @param array $data
     * @return \stdClass
     * @throws \Magento\Framework\Webapi\Exception
     */
    protected function checkout($data)
    {
      $transaction = new \BeGateway\GetPaymentToken;
      $helper = $this->getModuleHelper();

      $transaction->money->setAmount($data['order']['amount']);
      $transaction->money->setCurrency($data['order']['currency']);
      $transaction->setDescription($data['order']['description']);
      $transaction->setTrackingId($data['tracking_id']);
      $transaction->setLanguage($data['order']['language']);
      $transaction->customer->setFirstName(strval($data['order']['billing']->getFirstname()));
      $transaction->customer->setLastName(strval($data['order']['billing']->getLastname()));
      $transaction->customer->setAddress(strval($data['order']['billing']->getStreetLine(1)));
      $transaction->customer->setCity(strval($data['order']['billing']->getCity()));
      $transaction->customer->setCountry(strval($data['order']['billing']->getCountryId()));
      $transaction->customer->setZip($data['order']['billing']->getPostcode());
      $transaction->setTestMode(intval($this->getConfigHelper()->getTestMode()) == 1);

      if (in_array(strval($data['order']['billing']->getCountryId()), array('US', 'CA'))) {
        $billingAddress = $data['order']['billing']->getData();
        $region = $helper->getRegionFactory()->create()->load($billingAddress['region_id']);
        $transaction->customer->setState($region->getCode());
      }

      if (!empty(strval($data['order']['customer']['email']))) {
        $transaction->customer->setEmail(strval($data['order']['customer']['email']));
      }

      $transaction->customer->setPhone(strval($data['order']['billing']->getTelephone()));

      $notification_url = $data['urls']['notify'];
      $transaction->setNotificationUrl($notification_url);

      $transaction->setSuccessUrl($data['urls']['return_success']);
      $transaction->setDeclineUrl($data['urls']['return_failure']);
      $transaction->setFailUrl($data['urls']['return_failure']);
      $transaction->setCancelUrl($data['urls']['return_cancel']);

      $payment_methods = $this->getCheckoutPaymentMethodTypes();

      $trx_type = $this->getCheckoutTransactionTypes();

      if ($trx_type == $helper::AUTHORIZE) {
        $transaction->setAuthorizationTransactionType();
      }

      if (in_array($helper::CREDIT_CARD, $payment_methods)) {
        $cc = new \BeGateway\PaymentMethod\CreditCard;
        $transaction->addPaymentMethod($cc);
      }

      if (in_array($helper::CREDIT_CARD_HALVA, $payment_methods)) {
        $halva = new \BeGateway\PaymentMethod\CreditCardHalva;
        $transaction->addPaymentMethod($halva);
      }

      if (in_array($helper::ERIP, $payment_methods)) {
        $erip = new \BeGateway\PaymentMethod\Erip(array(
          'order_id' => $data['order']['increment_id'],
          'account_number' => strval($data['order']['increment_id']),
          'service_no' => $data['erip']['service_no'],
          'service_info' => array($data['erip']['service_info'])
        ));
        $transaction->addPaymentMethod($erip);
      }

      $transaction->additional_data->setMeta(array(
        'platform_data' => 'Magento v' . $helper->getMagentoVersion(),
        'integration_data' => 'BeGateway Magento 2 Module v' . $helper->getVersion()
      ));

      $this->_addDebugData('token_request', $transaction);
      $response = $transaction->submit();

      $this->_addDebugData('token_response', $response);
      return $response;
    }

    /**
     * Order Payment
     * @param \Magento\Payment\Model\InfoInterface $payment
     * @param float $amount
     * @return $this
     * @throws \Magento\Framework\Exception\LocalizedException
     */
    public function order(\Magento\Payment\Model\InfoInterface $payment, $amount)
    {
        $order = $payment->getOrder();

        $orderId = ltrim(
            $order->getIncrementId(),
            '0'
        );

        $data = [
            'tracking_id' =>
                $this->getModuleHelper()->genTransactionId(
                  $orderId
                ),
            'transaction_types' =>
                $this->getConfigHelper()->getTransactionTypes(),
            'order' => [
                'increment_id' => $orderId,
                'currency' => $order->getBaseCurrencyCode(),
                'language' => $this->getModuleHelper()->getLocale(),
                'amount' => $amount,
                'usage' => $this->getModuleHelper()->buildOrderUsage(),
                'description' => __('Order # %1 payment', $orderId),
                'customer' => [
                    'email' => $this->getCheckoutSession()->getQuote()->getCustomerEmail(),
                ],
                'billing' =>
                    $order->getBillingAddress(),
                'shipping' =>
                    $order->getShippingAddress()
            ],
            'erip' => [
              'service_no' => $this->getConfigHelper()->getValue('erip_service_no'),
              'service_info' => array(
                __('Order # %1 payment', $orderId),
                $this->getModuleHelper()->buildOrderDescriptionText(
                  $order
                ),
              ),
            ],
            'urls' => [
                'notify' =>
                    $this->getModuleHelper()->getNotificationUrl(
                        $this->getCode()
                    ),
                'return_success' =>
                    $this->getModuleHelper()->getReturnUrl(
                        $this->getCode(),
                        'success'
                    ),
                'return_cancel'  =>
                    $this->getModuleHelper()->getReturnUrl(
                        $this->getCode(),
                        'cancel'
                    ),
                'return_failure' =>
                    $this->getModuleHelper()->getReturnUrl(
                        $this->getCode(),
                        'failure'
                    ),
            ]
        ];

        $this->getConfigHelper()->initGatewayClient();

        try {
            $responseObject = $this->checkout($data);

            $isBeGatewaySuccessful =
                $responseObject->isSuccess() && !empty($responseObject->getRedirectUrl());

            if (!$isBeGatewaySuccessful) {
                $errorMessage = $responseObject->getMessage();

                $this->getCheckoutSession()->setBeGatewayLastCheckoutError(
                    $errorMessage
                );

                $this->getModuleHelper()->throwWebApiException($errorMessage);
            }

            $payment->setTransactionId($responseObject->getToken());
            $payment->setIsTransactionPending(true);
            $payment->setIsTransactionClosed(false);

            $this->getModuleHelper()->setPaymentTransactionAdditionalInfo(
                $payment,
                $responseObject
            );

            $this->getCheckoutSession()->setBeGatewayCheckoutRedirectUrl(
                $responseObject->getRedirectUrl()
            );

            $this->_writeDebugData();
            return $this;
        } catch (\Exception $e) {
            $this->_addDebugData('exception', $e->getMessage());

            $this->getCheckoutSession()->setBeGatewayLastCheckoutError(
                $e->getMessage()
            );

            $this->_writeDebugData();
            $this->getModuleHelper()->maskException($e);
        }
    }

    /**
     * Payment Capturing
     * @param \Magento\Payment\Model\InfoInterface $payment
     * @param float $amount
     * @return $this
     * @throws \Magento\Framework\Webapi\Exception
     */
    public function capture(\Magento\Payment\Model\InfoInterface $payment, $amount)
    {
        /** @var \Magento\Sales\Model\Order $order */
        $order = $payment->getOrder();

        $this->_addDebugData('capture_process', 'Capture transaction for order #' . $order->getIncrementId());

        $authTransaction = $this->getModuleHelper()->lookUpAuthorizationTransaction(
            $payment
        );

        if (!isset($authTransaction)) {
            $errorMessage = __('Capture transaction for order # %1 cannot be finished (No Authorize Transaction exists)',
                $order->getIncrementId()
            );

            $this->_addDebugData('capture_error', $errorMessage);
            $this->_writeDebugData();

            $this->getModuleHelper()->throwWebApiException(
                $errorMessage
            );
        }

        try {
            $this->doCapture($payment, $amount, $authTransaction);
        } catch (\Exception $e) {
            $this->_addDebugData('exception', $e->getMessage());
            $this->_writeDebugData();
            $this->getModuleHelper()->maskException($e);
        }
        $this->_writeDebugData();

        return $this;
    }

    /**
     * Payment refund
     *
     * @param \Magento\Payment\Model\InfoInterface $payment
     * @param float $amount
     * @return $this
     * @throws \Magento\Framework\Webapi\Exception
     */
    public function refund(\Magento\Payment\Model\InfoInterface $payment, $amount)
    {
        /** @var \Magento\Sales\Model\Order $order */
        $order = $payment->getOrder();

        $this->_addDebugData('refund_process', 'Refund transaction for order #' . $order->getIncrementId());

        $captureTransaction = $this->getModuleHelper()->lookUpCaptureTransaction(
            $payment
        );

        if (!isset($captureTransaction)) {
            $errorMessage = __('Refund transaction for order # %1 cannot be finished (No Capture Transaction exists)',
                $order->getIncrementId()
            );

            $this->_addDebugData('refund_error', $errorMessage);
            $this->_writeDebugData();

            $this->getMessageManager()->addError($errorMessage);

            $this->getModuleHelper()->throwWebApiException(
                $errorMessage
            );
        }

        try {
            $this->doRefund($payment, $amount, $captureTransaction);
        } catch (\Exception $e) {
            $this->_addDebugData('exception', $e->getMessage());
            $this->_writeDebugData();

            $this->getMessageManager()->addError(
                $e->getMessage()
            );

            $this->getModuleHelper()->maskException($e);
        }
        $this->_writeDebugData();

        return $this;
    }

    /**
     * Payment Cancel
     * @param \Magento\Payment\Model\InfoInterface $payment
     * @return $this
     */
    public function cancel(\Magento\Payment\Model\InfoInterface $payment)
    {
        $this->void($payment);

        return $this;
    }

    /**
     * Void Payment
     * @param \Magento\Payment\Model\InfoInterface $payment
     * @return $this
     * @throws \Magento\Framework\Webapi\Exception
     */
    public function void(\Magento\Payment\Model\InfoInterface $payment)
    {
        /** @var \Magento\Sales\Model\Order $order */

        $order = $payment->getOrder();

        $this->_addDebugData('void_process', 'Void transaction for order #' . $order->getIncrementId());

        $referenceTransaction = $this->getModuleHelper()->lookUpVoidReferenceTransaction(
            $payment
        );

        if ($referenceTransaction->getTxnType() == \Magento\Sales\Model\Order\Payment\Transaction::TYPE_AUTH) {
            $authTransaction = $referenceTransaction;
        } else {
            $authTransaction = $this->getModuleHelper()->lookUpAuthorizationTransaction(
                $payment
            );
        }

        if (!isset($authTransaction) || !isset($referenceTransaction)) {
            $errorMessage = __('Void transaction for order # %1 cannot be finished (No Authorize / Capture Transaction exists)',
                            $order->getIncrementId()
            );
            $this->_addDebugData('void_error', $errorMessage);
            $this->_writeDebugData();

            $this->getModuleHelper()->throwWebApiException($errorMessage);
        }

        try {
            $this->doVoid($payment, $authTransaction, $referenceTransaction);
        } catch (\Exception $e) {
            $this->_addDebugData('exception', $e->getMessage());
            $this->_writeDebugData();
            $this->getModuleHelper()->maskException($e);
        }
        $this->_writeDebugData();

        return $this;
    }

    /**
     * Determines method's availability based on config data and quote amount
     *
     * @param \Magento\Quote\Api\Data\CartInterface|null $quote
     * @return bool
     */
    public function isAvailable(\Magento\Quote\Api\Data\CartInterface $quote = null)
    {
        return parent::isAvailable($quote) &&
            $this->getConfigHelper()->isMethodAvailable();
    }

    /**
     * Checks base currency against the allowed currency
     *
     * @param string $currencyCode
     * @return bool
     */
    public function canUseForCurrency($currencyCode)
    {
        return $this->getModuleHelper()->isCurrencyAllowed(
            $this->getCode(),
            $currencyCode
        );
    }

    // returns array for \Magento\Payment\Model\Method\Logger $logger
    // @return array
    protected function _getDebugMessage() {
        return $this->_debugData;
      }
}
